##    Radio Free Pete -- streaming Internet jukebox for the masses
##    Copyright (C) 2001 Peter Gautier
##
##    This program is free software; you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation; either version 2 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program; if not, write to the Free Software
##    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


from Persistence import Persistent
import threading, os, sys, string
from pprint import pprint
from telnetlib import Telnet
from time import sleep
from urllib import quote

_thread = None

class Icecast(Persistent):
	""" Class for controlling the Icecast Server """

	def __init__(self):
		""" initialize a new Icecast object """
		pass
	
	def start_icecast(self, REQUEST=None):
		""" Start the Icecast process. This class uses the icecast config
		information from the ZRFP properties to feed to the thread that
		it runs Icecast in. """

		zrfp   = self.aq_parent
		errmsg = ''
		_v_thread = IcecastThread(zrfp['icecast_port'], zrfp['icecast_binary'],
								  zrfp['icecast_config_file'], zrfp['background_icecast'])
		_v_thread.start()
		while 1:
			_v_thread.is_ready.wait(0.05)
			_v_thread.has_error.wait(0.05)
			if _v_thread.is_ready.isSet():
				print "Icecast: Started thread"
				break
			elif _v_thread.has_error.isSet():
				print "Icecast: Couldn't start"
				errmsg = 'Error: Icecast could not be started.'
				break
		if REQUEST:
			if errmsg: query_string = 'errmsg=' + quote(errmsg)
			else     : query_string = None
			self.aq_parent._bounce(query_string)
		
	def is_running(self):
		""" Return true if Icecast is running and false if it is not """
		try:
			print "Icecast: Pinging Icecast server. Ignore the socket error if you see one."
			tn = Telnet(string.strip(self.icecast_host), self.icecast_port)
			tn.close()
			return 1
		except:
			return None

	def shutdown_icecast(self):
		""" Stop the Icecast server """
		tt = self.get_thread()
		zrfp = self.aq_parent
		tn = Telnet(zrfp['icecast_host'], int(zrfp['icecast_port']))
		tn.read_until('This is a bogus string', 1) # sleep a sec
		tn.write('ADMIN %s\n\n' % string.strip(zrfp['icecast_administrator_password']))
		tn.read_until('-> ', 1)
		tn.write('oper %s\n' % string.strip(zrfp['icecast_operator_password']))
		tn.read_until('> ', 1)
		tn.write('shutdown\n')
		tn.read_very_eager()
		tn.close() 
		print "Icecast: giving Icecast time to go away"
		try:
			while tt.isAlive():
				sleep(0.01)
		except AttributeError:
			# We must have started icecast in the background . . .
			sleep(8)
		self.aq_parent._bounce()

	def get_thread(self):
		global _thread
		return _thread
		
class IcecastThread(threading.Thread):
	""" Runs the Icecast server in a separate thread """

	def __init__(self, port, binary, configfile, background_icecast):
		""" initialize a new IcecastThread object """
		threading.Thread.__init__(self)
		self.port		        = port
		self.binary		        = binary
		self.configfile         = configfile
		self.is_ready           = threading.Event()
		self.has_error          = threading.Event()
		self.background_icecast = background_icecast
		
	def run(self):
		""" run the thread """
		if self.background_icecast:
			self.run_in_background()
		else:
			self.run_in_foreground()

	def run_in_foreground(self):
		global _thread
		_thread = self
		pipe = os.popen("%s -c %s -P %s" % (self.binary, self.configfile, self.port), 'r')
		line = pipe.readline()
		while line:
			if not self.is_ready.isSet() and string.find(line, 'Uptime:') >= 0 :
				print "IcecastThread: *** Ready ***"
				self.is_ready.set()
			print "IcecastThread: %s" % line[:-1]
			line = pipe.readline()
		if not self.is_ready.isSet():
			self.has_error.set()
		print "IcecastThread: Icecast is going away now."
		sys.exit(0)
				
	def run_in_background(self):
		os.system("%s -c %s -P %s -b" % (self.binary, self.configfile, self.port))
		sleep(2)
		self.is_ready.set()
		print "IcecastThread: *** Ready ***"
		print "IcecastThread: Icecast is running in the background."
		sys.exit(0)
