##	  Radio Free Pete -- streaming Internet jukebox for the masses
##	  Copyright (C) 2001 Peter Gautier
##
##	  This program is free software; you can redistribute it and/or modify
##	  it under the terms of the GNU General Public License as published by
##	  the Free Software Foundation; either version 2 of the License, or
##	  (at your option) any later version.
##
##	  This program is distributed in the hope that it will be useful,
##	  but WITHOUT ANY WARRANTY; without even the implied warranty of
##	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
##	  GNU General Public License for more details.
##
##	  You should have received a copy of the GNU General Public License
##	  along with this program; if not, write to the Free Software
##	  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA

# mp3 library
import mp3
# standard library modules
import os, string, re
from exceptions import Exception

class MP3Finder:
	""" Finds mp3 files and extracts the information we need from them """

	def __init__(self, mp3_basedir):
		self.stack = [mp3_basedir]
		self.files = []
		self.index = 0
		
	def __getitem__(self, index):
		while 1:
			try:
				file = self.files[self.index]
				self.index = self.index + 1
			except IndexError:
				# pop next directory from stack
				self.directory = self.stack.pop()
				self.files = os.listdir(self.directory)
				self.index = 0
			else:
				# got a filename
				fullname = os.path.join(self.directory, file)
				if os.path.isdir(fullname): #and not os.path.islink(fullname):
					self.stack.append(fullname)
					continue
				if string.find(file, '.mp3') >= 0:
					filename = fullname			
					if hasattr(self, 'cache') and self.cache.has_key(filename):
						#print "MP3Finder: skipping cached track %s" % filename
						continue
					try:
						artist, album, title, seconds, bitrate = self.get_mp3_info(filename)
					except ZRFPException, msg:
						print "MP3Finder: error %s for %s" % (msg, filename)
						continue
					return({'filename': filename,
							'short_filename': file,
							'artist': artist,
							'album': album,
							'title': title,
							'seconds': seconds,
							'bitrate': bitrate})
			
	def set_cache(self, cache):
		self.cache = cache

	def get_mp3_info(self, track):
		""" Get id3 tag and file header information """
		# this try/except is mostly for importing playlists from files
		try:
			id3tag = mp3.get_mp3tag(track)
		except OSError, msg:
			raise ZRFPException, '%s' % msg
		
		if not id3tag:
			raise ID3TagException, "Can't get id3 tag"
		else:
			artist, album, title = self.clean_tag(id3tag['ARTIST'], id3tag['ALBUM'], id3tag['TITLE'])

		try:
			mp3info = mp3.mp3info(track)
		except:
			raise MP3HeaderException, "Can't get mp3 header"

		seconds = (int(mp3info['MM']) * 60) + int(mp3info['SS'])
		bitrate = mp3info['BITRATE']
		return artist, album, title, seconds, bitrate

	def clean_tag(self, artist, album, title):
		""" strip undesirable characters and whitespace out of the id3 tags """
		# Kill anything following a null character (and the null)
		artist	= string.split(artist, "\0")[0]
		album	= string.split(album, "\0")[0]
		title	= string.split(title, "\0")[0]
		
		# Convert "_" to " "
		artist	= string.replace(artist, '_', ' ')
		album	= string.replace(album, '_', ' ')
		title	= string.replace(title, '_', ' ')
		
		# Strip leading "The " from artist
		artist	= re.sub(r'(?i)^\s*the\s+', '', artist)
		
		# Strip out characters we don't want (only allow the list below)
		patt = re.compile(r'[^\w\s\.\&\'\"\%\!\$\,\:\;\-\*\+\=\|\?}{><)(\[\]]')
		artist	= patt.sub('', artist)
		album	= patt.sub('', album)
		title	= patt.sub('', title)

		# Strip all leading and trailing whitespace
		artist	= string.strip(artist)
		album	= string.strip(album)
		title	= string.strip(title)

		if re.match(r'^\s*$', artist): raise ID3TagException, "No artist tag"
		if re.match(r'^\s*$', album) : raise ID3TagException, "No album tag"
		if re.match(r'^\s*$', title) : raise ID3TagException, "No title tag"
		
		return artist, album, title

class ZRFPException(Exception):
	"""Base class for all RFP Exceptions"""
	def __init__(self, args=None):
		self.args = args

class ID3TagException(ZRFPException):
	""" Raise for problems with reading id3 tag information """
	pass

class MP3HeaderException(ZRFPException):
	""" Raise for problems with reading mp3 header information """
	pass
