##    Radio Free Pete -- streaming Internet jukebox for the masses
##    Copyright (C) 2001 Peter Gautier
##
##    This program is free software; you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation; either version 2 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program; if not, write to the Free Software
##    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from Persistence import Persistent

rand_module  = None
get_rand_int = None
try:
	import devrandommodule
	rand_module = 'devrandommodule'
except ImportError:
	import whrandom
	rand_module = 'whrandom'

if rand_module == 'devrandommodule':
	def get_rand_int(ceil):
		return devrandommodule.strong_rand_int(ceil)
else:
	def get_rand_int(ceil):
		return whrandom.randint(0, ceil - 1)

class Playlist(Persistent):

	def __init__(self):
		self.tracks = {}

	def add(self, artist_id=None, album_id=None, track_id=None, start=1):
		""" router for add functions """
		if track_id: self.add_track(artist_id, album_id, track_id)
		elif album_id: self.add_album(artist_id, album_id)
		else: self.add_albums(artist_id)
		self.aq_parent._bounce()
		
	def add_albums(self, artist_id):
		for album_id in self.aq_parent.get_album_ids(artist_id):
			for track_id in map(lambda x: x.id, self.aq_parent.get_tracks(artist_id, album_id)):
				self.add_track(artist_id, album_id, track_id)

	def add_album(self, artist_id, album_id):
		for track_id in map(lambda x: x.id, self.aq_parent.get_tracks(artist_id, album_id)):
			self.add_track(artist_id, album_id, track_id)

	def add_track(self, artist_id=None, album_id=None, track_id=None):
		tt = self.tracks
		if not tt.has_key(artist_id): tt[artist_id] = {}
		if not tt[artist_id].has_key(album_id): tt[artist_id][album_id] = {}
		if not tt[artist_id][album_id].has_key(track_id):
			tt[artist_id][album_id][track_id] = None
		self.tracks = tt

	def remove(self, artist_id=None, album_id=None, track_id=None, start=1):
		""" remove a track or tracks from the playlist """
		try:
			if not album_id: self.remove_albums(artist_id)
			elif not track_id: self.remove_album(artist_id, album_id)
			else: self.remove_track(artist_id, album_id, track_id)
		except KeyError:
			pass # dumbass
		
		self.aq_parent._bounce()
		
	def remove_albums(self, artist_id=None):
		del(self.tracks[artist_id])
		self.tracks = self.tracks

	def remove_album(self, artist_id=None, album_id=None):
		del(self.tracks[artist_id][album_id])
		self.tracks = self.tracks

	def remove_track(self, artist_id=None, album_id=None, track_id=None):
		del(self.tracks[artist_id][album_id][track_id])

	def get_count_for_artist(self, artist_id):
		t = self.tracks
		if not t.has_key(artist_id):
			return 0

		ret = 0
		for album_id in t[artist_id].keys():
			ret = ret + len(t[artist_id][album_id].keys())
		return ret

	def is_in_playlist(self, artist_id, album_id, track_id):
		try:
			return self.tracks[artist_id][album_id].has_key(track_id)
		except KeyError:
			return 0

	def change_title(self, title):
		""" change the title """
		self.title=title
		self.aq_parent._bounce()

	def get_random_track(self):
		global get_rand_int
		dist       = {}
		tot_weight = 0
		for ar in self.tracks.keys():
			for al in self.tracks[ar].keys():
				for tr in self.tracks[ar][al].keys():
					weight = self.aq_parent.get_track(ar, al, tr).get_weight()
					tot_weight = tot_weight + weight
					dist[tot_weight] = (ar, al, tr)
		rand = get_rand_int(tot_weight)
		#print "Playlist: random is %s --- " % rand,
		while 1:
			if not dist.has_key(rand):
				rand = rand + 1
			else:
				break
		#print "got key %s" % rand
		return dist[rand]

	def get_track_struct(self):
		return self.tracks
