##	  Radio Free Pete -- streaming Internet jukebox for the masses
##	  Copyright (C) 2001 Peter Gautier
##
##	  This program is free software; you can redistribute it and/or modify
##	  it under the terms of the GNU General Public License as published by
##	  the Free Software Foundation; either version 2 of the License, or
##	  (at your option) any later version.
##
##	  This program is distributed in the hope that it will be useful,
##	  but WITHOUT ANY WARRANTY; without even the implied warranty of
##	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
##	  GNU General Public License for more details.
##
##	  You should have received a copy of the GNU General Public License
##	  along with this program; if not, write to the Free Software
##	  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA

import threading, sys, string, traceback, os, re
from ZPublisher import Client
import urllib
from telnetlib import Telnet
from pyshout import PyShout
from time import sleep, time
from exceptions import Exception
from string import upper, split, strip
from whrandom import choice
from mp3 import mp3info

_threads = {}
last_commercial = {}

class Stream:
	""" Class for controlling the streaming of data to the Icecast server """

	def __init__(self):
		self._v_thread = None
		self.requests  = []
		
	def start_stream(self, REQUEST=None):
		""" Start a thread responsible for streaming mp3s to the Icecast server """
		global last_commercial
		last_commercial[self.id] = time()
		pp = self.aq_parent
		if REQUEST: self.mp3_url = '%s/%s' % (self.REQUEST.URL1, 'get_next_mp3')
		try:
			self._v_thread = StreamThread(self.id, self.mp3_url, pp.icecast_host,
							  pp.icecast_port, pp.icecast_encoder_password)
			self._v_thread.start()
			self._v_thread.is_ready.wait()
			if REQUEST:
				self.aq_parent._bounce()
		except IOError:
			errmsg = "The most common source of this error is not starting Icecast before you try to start a stream. Please make sure that Icecast is running (especially if the specific error is 'Couldn't connect to remote host')"
			type, val, tb = sys.exc_info()
			errmsg = "The specific error is '%s'. %s" % (val, errmsg)
			del type, val, tb
			if REQUEST:
				self.aq_parent._bounce('errmsg=%s' % urllib.quote(errmsg))
			
	def get_thread(self):
		""" Return reference to out thread """
		try:
			global _threads
			return _threads[self.id]
		except KeyError:
			pass

	def thread_is_alive(self):
		""" Boolean test for whether or not we have a reference to a live StreamThread """
		t = self.get_thread()
		if t and t.isAlive():
			if not t.is_ready.isSet():
				t.is_ready.wait(10)
				if not t.is_ready.isSet():
					print "Stream: *** Thread for %s is undead!" % self.id
					_threads[self.id] = None
					print "Stream: *** Trying to restart stream"
					self.restart_thread()
					return self.thread_is_alive()
			return 1
		else:
			return 0
		
	def get_random_track(self):
		global get_rand_int
		pp		   = self.aq_parent
		tot_weight = 0
		dist	   = {}
		for artist in pp.Artists.objectIds(['Folder']):
			for album in pp.Artists._getOb(artist).objectIds(['Folder']):
				for track in pp.Artists._getOb(artist)._getOb(album).objectValues(['Track']):
					weight = track.get_weight()
					tot_weight = tot_weight + weight
					dist[tot_weight] = (artist, album, track.id)
		rand = get_rand_int(tot_weight)
		while 1:
			if not dist.has_key(rand):
				rand = rand + 1
			else:
				break
		return dist[rand]

	def get_next_mp3(self, startup=0):
		""" Get the current request or, if none, a random mp3 file """
		global last_commercial
		is_request = 0
		if not startup and int(self.commercials_interval):
			if int(time()) >= last_commercial[self.id] +\
			   (int(self.commercials_interval) * 60):
				last_commercial[self.id] = int(time())
				comm = self.get_commercial()
				if not comm:
					print "Stream: *** Problem with commercials. Skipping."
				else:
					return comm
		else:
			last_commercial[self.id] = int(time())

		if self.requests:
			tmp = self.requests
			info = tmp.pop(0)
			track = tuple(info)
			self.requests = tmp
			is_request = 1
		elif self.playlist:
			track = self.aq_parent._getOb(self.playlist).get_random_track()
		else:
			track = self.get_random_track()
		track = apply(self.aq_parent.get_track, track)
		track.num_plays = track.num_plays + 1
		if is_request: track.is_request = 1
		else		 : track.is_request = 0

		return '%s' % vars(track)

	def get_commercial(self):
		if os.path.isdir(self.aq_parent.commercials_directory):
			mp3s = filter(lambda x: re.search(r'\.mp3$', x), \
						  os.listdir(self.aq_parent.commercials_directory))
			if mp3s:
				file = choice(mp3s)
				filename = self.aq_parent.commercials_directory + '/' + file
				try:
					info = mp3info(filename)
					return {'artist': 'rfp', 'album': 'rfp', 'title': 'commercial',
							'seconds': (int(info['MM']) * 60) + int(info['SS']),
							'filename': filename, 'id': 'commercial', 'is_request': 0,
							'bitrate': info['BITRATE']}
				except:
					print "Stream: *** Couldn't get mp3 header info for commercial %s" % filename
					return

	def skip_song(self):
		""" Skip the currently playing song """
		np = self.get_now_playing()
		clean_name = self.aq_parent.clean_name
		tt = self.aq_parent.Artists._getOb(upper(clean_name(np['artist']))).\
			 _getOb(upper(clean_name(np['album']))).\
			 _getOb(np['id'])
		tt.num_skips = tt.num_skips + 1
		thread = self.get_thread()
		thread.is_ready.clear()
		thread.skip_event.set()
		thread.is_ready.wait()
		self._v_thread = thread
		self.aq_parent._bounce()
		
	def kill_stream(self):
		""" Kill the thread that is streaming mp3 data to the Icecast server """
		thread = self.get_thread()
		thread.bail_event.set()
		while thread.isAlive():
			sleep(0.05)
		self._v_thread = None
		self.aq_parent._bounce()

	def get_now_playing(self):
		""" Returns information about the currently playing track """
		tt = self.get_thread()
		if tt: return tt.now_playing

	def tune_in(self):
		""" Try to get the user's mp3 player to connect to our Icecast server """
		pp	  = self.aq_parent
		redir = 'http://%s:%s/%s' % (pp.icecast_host, pp.icecast_port, urllib.quote(self.id)) 
		self.REQUEST.RESPONSE.setHeader('content-type', 'audio/x-mpegurl')
		self.REQUEST.RESPONSE.setHeader('content-length', '%d' % (len(redir)+1))
		self.REQUEST.RESPONSE.write(redir + '\n')

	def change_title(self, title):
		""" change the title """
		self.title=title
		self.aq_parent._bounce()

	def change_playlist(self, playlist):
		""" change the playlist """
		self.playlist = playlist
		self.aq_parent._bounce()

	def request(self, artist_id, album_id, track_id):
		""" make a request """
		tt = self.aq_parent.Artists._getOb(artist_id)._getOb(album_id)._getOb(track_id)
		tt.num_requests = tt.num_requests + 1
		tmp = self.requests
		tmp.append([artist_id, album_id, track_id])
		self.requests = tmp
		self.aq_parent._bounce()
		
	def get_song_times(self):
		np = self.get_now_playing()
		total = np['seconds']
		tt = self.get_thread()
		elapsed = tt.secs_elapsed
		return '%s/%s' % (self.format_secs(elapsed), self.format_secs(total))

	def get_track_stats(self):
		np = self.get_now_playing()
		tt = self.aq_parent.get_track(np['artist'], np['album'], np['title'])
		ret = {'num_plays': tt.num_plays, 'num_requests': tt.num_requests,
			   'num_skips': tt.num_skips, 'num_ratings': tt.num_ratings,
			   'rating_total': tt.rating_total}
		if tt.num_ratings:
			ret.update({'rating': '%.2f' % (float(tt.rating_total) / float(tt.num_ratings))})
		return ret

	def rate_track(self, rating, artist, album, title):
		""" rate a track """
		tt = self.aq_parent.get_track(artist, album, title)
		tt.num_ratings = tt.num_ratings + 1
		tt.rating_total = tt.rating_total + rating
		self.aq_parent._bounce()

	def set_commercial_interval(self, minutes):
		""" Change the number of minutes between commercials """
		self.commercials_interval = minutes
		self.aq_parent._bounce()

	def toggle_skippable(self):
		""" Toggles the allow_skips property """
		if self.allow_skips: self.allow_skips = 0
		else			   : self.allow_skips = 1
		self.aq_parent._bounce()

	def restart_thread(self):
		""" uh """
		tn = Telnet(strip(self.aq_parent.icecast_host), int(self.aq_parent.icecast_port))
		tn.read_until('This is a bogus string', 1) # sleep a sec
		tn.write('ADMIN %s\n\n' % strip(self.aq_parent.icecast_administrator_password))
		tn.read_until('-> ', 1)
		tn.write('oper %s\n' % strip(self.aq_parent.icecast_operator_password))
		tn.read_until('> ', 1)
		tn.write('sources\n')
		ret = tn.read_until('> ', 1)
		lines = split(ret, "\n")
		lines.pop(0)
		iceid = None
		for line in lines[:-2]:
			if re.search(r'Mountpoint: /%s' % self.id, line):
				mm = re.search(r'Id: (\d+)', line)
				iceid = mm.group(1)
				break
		if iceid:
			tn.write('kick %s\n' % iceid)
			tn.read_until('> ', 1)
		tn.close()
		sleep(3)
		self.start_stream(REQUEST=None)
		
class StreamThread(threading.Thread):
	""" Class to implement the thread that sends mp3 data to icecast """
	
	def __init__(self, stream_id, mp3_url, icecast_host, icecast_port,
			 icecast_password):
		threading.Thread.__init__(self)
		self.stream_id	 = stream_id
		self.mp3_url	 = mp3_url
		self.icecast_host = icecast_host
		self.icecast_port = icecast_port
		self.icecast_password = icecast_password
		self.pyshout	 = self.get_handle()
		self.skip_event = threading.Event()
		self.bail_event = threading.Event()
		self.is_ready	= threading.Event()
		self.secs_elapsed = 0
		self.now_playing = None
		
	def run(self):
		print "StreamThread started"
		global last_commercial
		last_commercial[self.stream_id] = time()
		try:
			self.scream()
		except NormalStreamThreadExitException:
			print "%s says goodbye" % self.stream_id
		except:
			type, value, tb = sys.exc_info()
			print "StreamThread: Exception [ %s: %s ]" % (type, value)
			print "StreamThread: \n\t[\n\t\t%s\n\t]" % string.join(traceback.format_tb(tb), "\n\t\t")
			print "StreamThread %s is going away now." % self
			global _threads, last_commercial
			try:
				del(_threads[self.stream_id])
				del(last_commercial[self.stream_id])
			except KeyError:
				pass
			sys.exit(0)

	def get_handle(self):
		""" returns a PyShout connection to the Icecast server """
		#print "StreamThread: Getting pyshout handle"
		psh = PyShout(self.icecast_host, int(self.icecast_port),
				  self.icecast_password)
		psh.set_bitrate(128000)
		psh.set_public(0)
		psh.set_name("Radio Free Pete")
		psh.set_description("Music")
		psh.set_genre("Whatever")
		psh.set_mount(urllib.quote(self.stream_id))
		psh.connect()
		return psh

	def get_next_mp3(self, startup=0):
		url = self.mp3_url + '?'
		if startup:	 url = url + '&startup=1'
		ret = Client.call(url)[1]
		try:
			track = eval(ret)
			return track
		except:
			print ret
			raise
		
	def scream(self):
		global _threads
		fuckups = 0
		_threads[self.stream_id] = self
		#print "StreamThread: _threads[%s] set to %s" % (self.stream_id,_threads[self.stream_id])
		mp3file = self.get_next_mp3(startup=1)
		while mp3file:
			#print "StreamThread: setting bitrate to %s" % mp3file['bitrate']
			self.pyshout.set_bitrate(mp3file['bitrate'])
			self.now_playing  = mp3file
			current_song_info = "%s by %s on %s" % (mp3file['title'], mp3file['artist'],
													mp3file['album'])
			if mp3file['is_request']:
				current_song_info = 'by request: ' + current_song_info
			print "StreamThread: Playing %s" % current_song_info
			total = 0
			try:
				f = open(mp3file['filename'], 'r')
				if not f:
					fuckups = fuckups + 1
					if fuckups > 10:
						raise ZRFPException, "Failed to get a good filehandle multiple times"
					else:
						raise IOError, "Opening file %s returned nothing" % mp3file['filename']
				else:
					fuckups = 0
					
			except IOError, msg:
				print "StreamThread: Error: %s" % msg
				mp3file = self.get_next_mp3()

			if not mp3file['title'] == 'commercial':
				urllib.urlopen('http://%s:%s/admin.cgi?mode=updinfo&pass=%s&mount=%s&song=%s' %
							   (strip(self.icecast_host),
								self.icecast_port,
								strip(self.icecast_password),
								urllib.quote(self.stream_id),
								urllib.quote(current_song_info)))
			self.is_ready.set()
			_threads[self.stream_id] = self
			nbuf = f.read(4096)
			t0 = int(time())
			while 1:
				self.secs_elapsed = int(time()) - t0
				self.bail_event.wait(0.001)
				if self.bail_event.isSet():
					print "StreamThread: Got a bail event"
					break

				self.skip_event.wait(0.001)
				if self.skip_event.isSet():
					if mp3file['title'] == 'commercial':
						self.skip_event.clear()
					else:
						print "StreamThread: Got a skip event"
						self.skip_event.clear()
						break

				buf	 = nbuf
				nbuf = f.read(4096)
				total = total + len(nbuf)

				if len(buf) == 0:
					break

				try:
					self.pyshout.send_data(buf)
					self.pyshout.sleep()
				except IOError:
					print "StreamThread: IOError in sending data. Reconnecting."
					self.pyshout = self.get_handle()
					break
			f.close()
			self.is_ready.clear()
			
			if self.bail_event.isSet():
				self.bail_event.clear()
				break

			mp3file = self.get_next_mp3()
				
		self.pyshout.disconnect()
		print "StreamThread: Exiting"
		raise NormalStreamThreadExitException, "StreamThread is exiting gracefully."
			
class ZRFPException(Exception):
	"""Base class for all RFP Exceptions"""
	def __init__(self, args=None):
		self.args = args

class NormalStreamThreadExitException(ZRFPException):
	""" Raise when StreamThread gets a bail event """
	pass

rand_module	 = None
get_rand_int = None
try:
	import devrandommodule
	rand_module = 'devrandommodule'
except ImportError:
	import whrandom
	rand_module = 'whrandom'

if rand_module == 'devrandommodule':
	def get_rand_int(ceil):
		return devrandommodule.strong_rand_int(ceil)
else:
	def get_rand_int(ceil):
		return whrandom.randint(0, ceil - 1)

