##	  Radio Free Pete -- streaming Internet jukebox for the masses
##	  Copyright (C) 2001 Peter Gautier
##
##	  This program is free software; you can redistribute it and/or modify
##	  it under the terms of the GNU General Public License as published by
##	  the Free Software Foundation; either version 2 of the License, or
##	  (at your option) any later version.
##
##	  This program is distributed in the hope that it will be useful,
##	  but WITHOUT ANY WARRANTY; without even the implied warranty of
##	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
##	  GNU General Public License for more details.
##
##	  You should have received a copy of the GNU General Public License
##	  along with this program; if not, write to the Free Software
##	  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA

# Zope includes
from Persistence import Persistent
# ZRFP includes
from MP3Finder import MP3Finder
# Standard library stuff
from pprint import pprint
from urllib import quote
from string import upper
import re, sys, string, os

uppercase = upper

class ZRFP(Persistent):
	""" Class that extends the ZRFP ZClass """

	def __init__(self):
		""" Add the folders to the Zope tree """
		self.manage_addFolder('Artists', 'Artists')

	def prune_tracks(self, num_artists, num_albums, num_tracks):
		for artist in self.Artists.objectIds(['Folder']):
			for album in self.get_album_ids(artist):
				for track in self.get_tracks(artist, album):
					if not self.file_exists(track.filename):
						# Man, too bad I have to stop indenting here!
						print "ZRFP: pruning track %s-%s-%s" % (artist, album, track.id)
						# I'm starting to think I shouldn't use _getOb and _delOb.
						# Close enough for jazz (for now)
						self.Artists._getOb(artist)._getOb(album).manage_delObjects(track.id)
						num_tracks = num_tracks - 1
				if not self.get_tracks(artist, album):
					self.Artists._getOb(artist).manage_delObjects(album)
					num_albums = num_albums - 1
			if not self.get_album_ids(artist):
				self.Artists.manage_delObjects(artist)
				num_artists = num_artists - 1
		return num_artists, num_albums, num_tracks

	def file_exists(self, filename):
		if os.path.exists(filename):
			return 1
		else:
			return None

	def set_track_nums(self, num_artists, num_albums, num_tracks):
		self.propertysheets.Static_properties.manage_changeProperties({'num_tracks': num_tracks,
																	   'num_artists': num_artists,
																	   'num_albums': num_albums})
		
	def crawl_tracks(self, reindex=0, add_only=0, remove_only=0, synch=0):
		""" Set the list of tagged mp3s found in the mp3_base_directory """
		cache = {}		
		num_artists = self.propertysheets.Static_properties.num_artists
		num_albums = self.propertysheets.Static_properties.num_albums
		num_tracks = self.propertysheets.Static_properties.num_tracks
		crawler = MP3Finder(self.mp3_base_directory)
		if reindex:
			self._clear_tracks()
			num_artists = 0
			num_albums = 0
			num_tracks = 0
		elif remove_only:
			(num_artists, num_albums, num_tracks) = \
						  self.prune_tracks(num_artists, num_albums, num_tracks)
			self.set_track_nums(num_artists, num_albums, num_tracks)
			self._bounce()
			return
		elif synch:
			(num_artists, num_albums, num_tracks) = \
						  self.prune_tracks(num_artists, num_albums, num_tracks)
			crawler.set_cache(self.get_filename_cache())
			# Fall through to crawling function. Basically go on to add_only.
		elif add_only:
			crawler.set_cache(self.get_filename_cache())
		print "ZRFP: crawling tracks in %s" % self.mp3_base_directory
		for mp3 in crawler:
			artist_id = uppercase(self.clean_name(mp3['artist']))
			album_id  = uppercase(self.clean_name(mp3['album']))
			try:
				if not artist_id in self.Artists.objectIds(['Folder']):
					num_artists = num_artists + 1
					self.Artists.manage_addFolder(artist_id, mp3['artist'])
				if not album_id in self.Artists._getOb(artist_id).objectIds(['Folder']):
					num_albums = num_albums + 1
					self.Artists._getOb(artist_id).manage_addFolder(album_id, mp3['album'])
				self.add_Track_object(artist_id, album_id, mp3)
				num_tracks = num_tracks + 1
			except:
				print "*** ZRFP Error: Can't add %s - %s - %s\n\t%s: %s" % (artist_id, album_id,
																			mp3['title'],
																			sys.exc_info()[0],
																			sys.exc_info()[1])
				continue

		del(crawler)
		self.set_track_nums(num_artists, num_albums, num_tracks)
		self._bounce()

	def _clear_tracks(self):
		self.manage_delObjects('Artists')
		self.manage_addFolder('Artists', 'Artists')
		self.propertysheets.Static_properties.manage_changeProperties({'num_tracks': 0,
																	   'num_artists': 0,
																	   'num_albums': 0})


	def start_all(self):
		""" This is just for my convenience in debugging. """
		ice = self.objectValues(['Icecast'])[0]
		ice.start_icecast(REQUEST=None)
		for stream in self.objectValues(['Stream']):
			stream.mp3_url = self.REQUEST.URL1 + '/' + stream.id + '/get_next_mp3'
			stream.start_stream(stream.playlist, REQUEST=None)

	def add_icecast_object(self):
		""" Add the Icecast object to the Zope tree """
		id = '%s_icecast' % self.id
		newob = self.Control_Panel.Products.ZRFP.ZRFP_Icecast()
		newob._setId(id)
		self._setObject(id, newob)
		print "ZRFP: Icecast object added"

	def add_Track_object(self, artist_id, album_id, mp3):
		""" Add the track object to the Zope tree """
		id = uppercase(self.clean_name(mp3['title']))
		newob = self.Control_Panel.Products.ZRFP.ZRFP_Track()
		newob._setId(id)
		newob.propertysheets.Track_properties.manage_changeProperties(mp3)
		self.Artists._getOb(artist_id)._getOb(album_id)._setObject(id, newob)
	
	def clean_name(self, name):
		""" Turn "dirty" names into Zope-friendly ones """
		name = re.sub(r'[^\w\d_]', '_', name)
		# Strip leading and trailing '_' chars
		name = re.sub(r'^_*', '', name)
		name = re.sub(r'_*$', '', name)
		return name

	def get_total_for_artist(self, artist_id):
		ret = 0
		for album in self.Artists._getOb(artist_id).objectValues(['Folder']):
			ret = ret + len(album.objectValues(['Track']))
		return ret

	def get_filename_cache(self):
		ret = {}
		for artist in self.Artists.objectIds(['Folder']):
			ret[artist] = {}
			for album in self.get_album_ids(artist):
				ret[artist][album] = {}
				for track in self.get_tracks(artist, album):
					ret[track.filename] = None
		return ret

	def get_album_ids(self, artist_id):
		return self.Artists._getOb(artist_id).objectIds(['Folder'])

	def get_title(self, artist_id=None, album_id=None, track_id=None):
		if track_id:
			return self.Artists._getOb(artist_id)._getOb(album_id)._getOb(track_id).title
		elif album_id:
			return self.Artists._getOb(artist_id)._getOb(album_id).title
		else:
			return self.Artists._getOb(artist_id).title

	def get_seconds_for_track(self, artist_id, album_id, track_id):
		return self.get_track(artist_id, album_id, track_id).seconds
		
	def get_tracks(self, artist_id, album_id):
		return self.Artists._getOb(artist_id)._getOb(album_id).objectValues(['Track'])

	def get_track(self, artist_id, album_id, track_id):
		clean_name = self.clean_name
		return self.Artists._getOb(upper(clean_name(artist_id))).\
			   _getOb(upper(clean_name(album_id))).\
			   _getOb(upper(clean_name(track_id)))

	def _bounce(self, query_string=''):
		dest = self.REQUEST.environ['HTTP_REFERER']
		# eliminate duplicate errors
		if query_string:
			if string.find(dest, query_string) >= 0:
				pass
			else:
				if string.find(dest, '?') >= 0:
					dest = dest + '&' + query_string
				else:
					dest = dest + '?' + query_string
		self.REQUEST.RESPONSE.redirect(dest)

	def make_playlist(self, newid=None):
		""" playlist constructor """
		newob = self.Control_Panel.Products.ZRFP.ZRFP_Playlist()
		if not self.is_valid_id(newid):
			self._bounce(query_string='errmsg=%s' % \
						 quote('Error: Please use only letters, digits, and underscores for the ID.'))
			return
		newob._setId(newid)
		self._setObject(newid, newob)
		self.REQUEST.RESPONSE.redirect(self.REQUEST.URL2+'/manage_workspace')

	def make_stream(self, newid=None):
		""" stream constructor """
		newob = self.Control_Panel.Products.ZRFP.ZRFP_Stream()
		if not self.is_valid_id(newid):
			self._bounce(query_string='errmsg=%s' % \
						 quote('Error: Please use only letters, digits, and underscores for the ID.'))
			return
		newob._setId(newid)
		self._setObject(newid, newob)
		self.REQUEST.RESPONSE.redirect(self.REQUEST.URL2+'/manage_workspace')

	def is_valid_id(self, id):
		if re.search(r'\W', id): return 0
		else				   : return 1

	def format_secs(self, secs):
		if secs / (60 * 60):
			hrs	 = secs / (60 * 60)
			secs = secs % (60 * 60)
			hrs = '%s:' % hrs
		else:
			hrs = ''
		mins = secs / 60
		secs = secs % 60
		return '%s%0.2d:%0.2d' % (hrs, mins, secs)
